from abc import abstractmethod
from typing import Dict, List, Tuple

from ..device_base import DeviceBase

import numpy as np


class Spectrometer(DeviceBase):
    @abstractmethod
    def get_spectra(self) -> Tuple[float, Dict[int, List[float]]]:
        """Timestamp and dict of spectral data for all boards."""
        ...

    def calc_tp(self, data: Dict[int, Tuple[float]], tp_range: List[int]) -> dict:
        tp_dict = {}
        if tp_range:
            for board_id, spectral_data in data.items():
                tp_sum = 0
                spectral_data = list(spectral_data)
                for i in range(0, len(tp_range), 2):
                    sum_in_range = np.nansum(
                        spectral_data[tp_range[i] : tp_range[i + 1]]
                    )
                    tp_sum += sum_in_range
                tp_dict[board_id] = [np.float32(tp_sum)]
        else:
            for board_id, spectral_data in data.items():
                tp = np.nansum(spectral_data)
                tp_list = [np.float32(tp)]
                tp_dict[board_id] = tp_list
        return tp_dict



from abc import abstractmethod

from ..device_base import DeviceBase


class DAConverter(DeviceBase):

    Manufacturer: str = ""
    Model: str

    @abstractmethod
    def set_voltage(self, mV: float, id: str) -> None: ...

    @abstractmethod
    def apply_voltage(self) -> None: ...

    @abstractmethod
    def finalize(self) -> None: ...

    @abstractmethod
    def close(self) -> None: ...



import ogameasure

from ... import get_logger
from ...core.security import busy
from ...core.units import dBm
from .power_meter_base import PowerMeter


class ML2437A(PowerMeter):
    """PowerMeter, which can measure IF signal power.

    Notes
    -----

    Configuration items for this device:

    host : str
        IP address for GPIB communicator.

    port : int
        GPIB port of using devices. Please check device setting.

    """

    Model = "ML2437A"
    Manufacturer = "Anritsu"

    Identifier = "host"

    def __init__(self):
        self.logger = get_logger(self.__class__.__name__)

        host = self.Config.host
        gpibport = self.Config.port
        com = ogameasure.gpib_prologix(host, gpibport)
        self.pm = ogameasure.Anritsu.ml2437a(com)

    def get_power(self) -> dBm:
        with busy(self, "busy"):
            power = self.pm.measure()
            return power * dBm

    def finalize(self) -> None:
        self.pm.com.close()
        return

    def close(self) -> None:
        self.finalize()



from .ccd_controller_base import CCDController


class CCDControllerSimulator(CCDController):
    Manufacturer: str = ""
    Model: str = ""
    Identifier = ""
    is_simulator = True

    def __init__(self) -> None:
        pass

    def capture(self, savepath: str) -> None:
        pass

    def finalize(self) -> None:
        pass

    def close(self) -> None:
        pass



import time

from ...core.security import busy
from .attenuator_base import CurrentAttenuator


class CPZ340516(CurrentAttenuator):
    """LOattenuator, which can convert by 8 channels.

    Notes
    -----
    Configuration items for this device:

    channel : Dict[str, int]
        Human-readable channel name. The value should be mapping from human readable
        version (str) to device level identifier (int). You can assign any name to the
        channels. For example, {100GHz = 1}

    range : "DA0_100mA" or "DA0_1mA"
        If DA0_100mA, 1-100mA is possible. Else if DA0_1mA, 1-10mA is possible.
        Default value is DA0_100mA.

    rate : float
        Default value is 0.1.

    rsw_id : {0, 1, ..., 16} or {"0", "1", ..., "9", "A", ..., "F"}
        Board identifier. This should be set to the same value as the rotary switch
        "RSW1" mounted on the side of the board. The board is shipped with default RSW1
        setting of 0. This ID would be non-zero, when multiple PCI board of same model
        are mounted on a single FA (Factory Automation) controller.

    See defaults setting file in ``neclib/defaults/config.toml``.

    """

    Manufacturer = "Interface"
    Model = "CPZ340516"

    Identifier = "rsw_id"

    def __init__(self):
        import pyinterface

        self.rsw_id = self.Config.rsw_id
        self.range = self.Config.range
        self.channel = self.Config.channel
        self.param_buff = {i: 0.0 for i in range(1, 9)}  # All in [mA]
        self.io = pyinterface.open(3405, self.rsw_id)
        for i in self.channel.values():
            try:
                self.io.set_outputrange(i, self.range)
            except Exception as e:
                if e == "lspci: Unable to load libkmod resources: error -2":
                    pass
                else:
                    raise ValueError(f"Invalid channel: {i}")

    def get_outputrange(self, id: int) -> dict:
        ch = self.Config.channel[id]
        with busy(self, "busy"):
            return self.io.get_outputrange(ch)

    def set_current(self, mA: float, id: str) -> None:
        ch = self.Config.channel[id]
        if ch not in self.param_buff.keys():
            raise ValueError(f"Invaild channel {ch}")
        else:
            self.param_buff[ch] = mA

    def set_outputrange(self, id: int, outputrange: str) -> None:
        self.io.set_outputrange(id, outputrange)

    def apply_current(self) -> None:
        with busy(self, "busy"):
            for i in range(0, 8):
                ch = int(list(self.param_buff.keys())[i])
                current = list(self.param_buff.values())[i]
                self.io.output_current(ch, current)
                time.sleep(0.001)

    def finalize(self) -> None:
        self.io.finalize()
        self.param_buff = {i: 0.0 for i in range(1, 9)}

    def close(self) -> None:
        pass



from .azd_ad import AZD_AD  # noqa: F401
from .cpz2724 import CPZ2724  # noqa: F401
from .cpz7204 import CPZ7204  # noqa: F401
from .cpz7415v import CPZ7415V  # noqa: F401
from .simulator import MotorSimulator  # noqa: F401



import astropy.units as u
import ogameasure

from ... import get_logger
from ...core.security import busy
from .vacuum_gauge_base import VacuumGauge


class TPG261(VacuumGauge):
    """Vacuum Gauge, which can check preassure of cryostat.

    Notes
    -----

    Configuration items for this device:

    communicator : str
    Communicator of thermometer. LAN or RS232 can be chosen.

    host : str
        IP address for ethernet communicator.
        If you use LAN communicator, you must set this parameter.

    port : int
        ethernet port of using devices. Please check device setting.
        If you use LAN communicator, you must set this parameter.

    rs232_port : str
        RS232 port of using devices.
        If you use RS232 communicator, you must set this parameter.

    """

    Model = "TPG261"
    Manufacturer = "Pfeiffer"

    Identifier = "communicator"

    def __init__(self):
        self.logger = get_logger(self.__class__.__name__)

        if self.Config.communicator == "LAN":
            com = ogameasure.ethernet(self.Config.host, self.Config.port)
            self.io = ogameasure.Pfeiffer.tpg261_lan(com)
            self.io.pres_unit_torr()

        elif self.Config.communicator == "RS232":
            com = ogameasure.serial(self.Config.rs232_port)
            self.io = ogameasure.Pfeiffer.tpg261(com)
            self.io.unit_command(unit=1)

        else:
            self.logger.warning(
                f"There is not exsited communicator: {self.Config.communicator}."
                "Please choose RS232 or LAN."
            )

    def get_pressure(self) -> u.Quantity:
        with busy(self, "busy"):
            if self.Config.communicator == "LAN":
                return self.io.pressure() * u.torr
            elif self.Config.communicator == "RS232":
                res = self.io.read_single_pressure()
                return res["value"] * u.torr

    def finalize(self) -> None:
        self.io.com.close()

    def close(self) -> None:
        self.io.com.close()



"""General file operations.

This module provides general file operations, not limited to a specific file format or
method to access the file system.

"""

__all__ = ["read"]

import os
from pathlib import Path
from typing import IO, Union
from urllib.parse import urlparse
from urllib.request import urlopen


def read(__path: Union[os.PathLike, str, IO], /, *, allow_remote: bool = True) -> str:
    """Read a file, if it's accessible via any available way.

    Parameters
    ----------
    __path
        The file path.
    allow_remote
        If True the file can be fetched from remote host, otherwise the look-up is
        limited to local file system.

    Returns
    -------
    content
        The file content.

    Raises
    ------
    FileNotFoundError
        If the file couldn't be found.
    URLError
        If the protocol (http, https, ftp, etc.) is specified in ``__path`` but invalid.

    Examples
    --------
    >>> neclib.core.files.read("relative/path/to/local-file.txt")
    >>> neclib.core.files.read("/absolute/path/to/local-file.txt")
    >>> neclib.core.files.read("https://example.com/file.txt")
    >>> neclib.core.files.read("http://192.168.xxx.xxx:80/file.txt")

    """
    if (not isinstance(__path, (str, os.PathLike))) and hasattr(__path, "read"):
        return __path.read()

    name = str(__path)
    _parsed = urlparse(name)
    scheme, path = _parsed.scheme, _parsed.path
    if scheme not in ("", "file"):
        if not allow_remote:
            raise FileNotFoundError(
                f"{name!r} appears to be a remote file, "
                "but remote access isn't allowed."
            )
        with urlopen(name) as response:
            return response.read().decode("utf-8")

    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"{name!r} does not exist.")
    try:
        return path.read_text()
    except UnicodeDecodeError:
        return path.read_bytes().decode("utf-8")



__all__ = ["CPZ2724"]

import struct
import time

from astropy import units as u

from ... import get_logger, utils
from .motor_base import Motor


class CPZ2724(Motor):
    """Digital I/O board for membrane opener of NANTEN2.

    Notes
    -----
    Configuration items for this device:

    rsw_id : {0, 1, ..., 16} or {"0", "1", ..., "9", "A", ..., "F"}
        Board identifier. This should be set to the same value as the rotary switch
        "RSW1" mounted on the side of the board. The board is shipped with default
        RSW1 setting of 0. This ID would be non-zero, when multiple PCI board of same
        model are mounted on a single FA (Factory Automation) controller.

    See defaults setting file in neclib/defaults/config.toml.
    """

    Manufacturer = "Interface"
    Model = "CPZ2724"

    Identifier = "rsw_id"

    def __init__(self) -> None:
        self.logger = get_logger(self.__class__.__name__)
        self.rsw_id = self.Config.rsw_id

        self.speed_to_rate = float((7 / 12) * 10000)

        self.io = self._initialize_io()

    @utils.skip_on_simulator
    def _initialize_io(self):
        import pyinterface

        io = pyinterface.open(2724, self.rsw_id)
        if io is None:
            raise RuntimeError("Cannot communicate with the CPZ board.")

        io.initialize()

        return io

    # Antenna Control

    def set_speed(
        self,
        speed: float,
        axis: str,
    ):
        self.max_rate = self.Config.max_rate
        speed_float = float(speed) * self.speed_to_rate

        if speed_float > self.max_rate:
            speed_float = self.max_rate
        if speed_float < -self.max_rate:
            speed_float = self.max_rate

        self.antenna_move(int(speed_float), axis)
        return

    def get_speed(self, axis: str) -> float:
        word = None
        if axis == "az":
            word = "IN1_16"
        elif axis == "el":
            word = "IN17_32"
        else:
            raise ValueError(f"No valid axis : {axis}")
        status = self.io.input_word(word)
        status_int = struct.unpack("<h", status.bytes)
        (speed,) = status_int
        speed_deg = speed / 3600
        return speed_deg * u.deg / u.s

    def antenna_move(self, speed: int, axis: str) -> None:
        word = None
        if axis == "az":
            word = "OUT1_16"
        elif axis == "el":
            word = "OUT17_32"
        else:
            raise ValueError(f"No valid axis : {axis}")

        cmd = list(
            map(
                int, "".join([format(b, "08b")[::-1] for b in struct.pack("<h", speed)])
            )
        )  # [::-1] for little endian.
        self.io.output_word(word, cmd)

    def antenna_stop(self):
        self.antenna_move(0, "az")
        self.antenna_move(0, "el")

    def antenna_status(self) -> dict[str, str]:
        status_dict = {}
        for i in ["az", "el"]:
            speed = self.get_speed(i)
            if speed == 0:
                antenna_status = "STOP"
            else:
                antenna_status = "MOVE"
            status_dict[i] = antenna_status
        return status_dict

    def set_step(self, step: int, axis: str) -> None:
        pass

    def get_step(self, axis: str) -> int:
        return 0

    # Dome Control

    def dome_move(self, speed: str, turn: str):
        buffer = [0, 1, 0, 0]
        if turn == "right":
            buffer[0] = 0
        elif turn == "left":
            buffer[0] = 1
        else:
            raise ValueError(f"No valid turn direction : {turn}")
        if speed == "low":
            buffer[2:4] = [0, 0]
        elif speed == "mid":
            buffer[2:4] = [1, 0]
        elif speed == "high":
            buffer[2:4] = [0, 1]
        else:
            raise ValueError(f"No valid speed : {speed}")
        self.io.output_point(buffer, 1)
        return

    def dome_oc(self, pos: str) -> None:
        # posにはopen or close を入れる
        ret = self.dome_status()
        if (ret[1].lower() != pos) & (ret[3].lower() != pos):
            buff = self.Config.position[pos.lower()]
            self.io.output_point(buff, 5)
        return

    def dome_stop(self) -> None:
        buff = [0]
        self.io.output_point(buff, 2)
        return

    def dome_pose(self) -> None:
        buff = [0, 0]
        self.io.output_point(buff, 5)
        return

    def dome_fan(self, fan: str) -> None:
        # fanにはon or off を入れる
        if fan == "on":
            fan_bit = [1, 1]
            self.io.output_point(fan_bit, 9)
        else:
            fan_bit = [0, 0]
            self.io.output_point(fan_bit, 9)

    def dome_status(self):
        ret = self.io.input_point(2, 6)
        # dome 開閉中："DRIVE", domeの扉止まっている："OFF"
        if ret[0] == 1:
            self.right_act = "DRIVE"
            self.right_pos = "MOVE"
        else:
            self.right_act = "OFF"
            if (ret[1] == 1) & (ret[2] == 0):
                self.right_pos = "OPEN"
            elif (ret[1] == 0) & (ret[2] == 1):
                self.right_pos = "CLOSE"
            else:
                self.right_pos = "POSE"

        if ret[3] == 1:
            self.left_act = "DRIVE"
            self.left_pos = "MOVE"
        else:
            self.left_act = "OFF"
            if (ret[4] == 1) & (ret[5] == 0):
                self.left_pos = "OPEN"
            elif (ret[4] == 0) & (ret[5] == 1):
                self.left_pos = "CLOSE"
            else:
                self.left_pos = "POSE"

        return [self.right_act, self.right_pos, self.left_act, self.left_pos]

    def dome_limit_check(self):
        limit = self.io.input_point(12, 4)
        ret = 0
        if limit[0:4] == [0, 0, 0, 0]:
            ret = 0
        elif limit[0:4] == [1, 0, 0, 0]:
            ret = 1
        elif limit[0:4] == [0, 1, 0, 0]:
            ret = 2
        elif limit[0:4] == [1, 1, 0, 0]:
            ret = 3
        elif limit[0:4] == [0, 0, 1, 0]:
            ret = 4
        elif limit[0:4] == [1, 0, 1, 0]:
            ret = 5
        elif limit[0:4] == [0, 1, 1, 0]:
            ret = 6
        elif limit[0:4] == [1, 1, 1, 0]:
            ret = 7
        elif limit[0:4] == [0, 0, 0, 1]:
            ret = 8
        elif limit[0:4] == [1, 0, 0, 1]:
            ret = 9
        elif limit[0:4] == [0, 1, 0, 1]:
            ret = 10
        elif limit[0:4] == [1, 1, 0, 1]:
            ret = 11
        elif limit[0:4] == [0, 0, 1, 1]:
            ret = 12
        return ret

    # Membrane Control

    def memb_oc(self, pos: str) -> None:
        # posには open or close を入れる
        ret = self.memb_status()
        if ret[1].lower() != pos:
            buff = self.Config.position[pos.lower()]
            self.io.output_point(buff, 7)
        return

    def memb_pose(self) -> None:
        buff = [0, 0]
        self.io.output_point(buff, 7)
        return

    def memb_status(self) -> list[str, str]:
        ret = self.io.input_point(8, 3)
        if ret[0] == 0:
            self.memb_act = "OFF"
            if (ret[1] == 0) & (ret[2] == 0):
                self.memb_pos = "POSE"
            elif (ret[1] == 1) & (ret[2] == 0):
                self.memb_pos = "OPEN"
            elif (ret[1] == 0) & (ret[2] == 1):
                self.memb_pos = "CLOSE"
        else:
            self.memb_act = "DRIVE"
            self.memb_pos = "MOVE"
        return [self.memb_act, self.memb_pos]

    # M2 Control

    def m2_move(self, dist: int):
        status = self.m2_status()
        puls = self.um_to_puls(dist, status)
        if puls is None:
            raise ValueError("Please change distance.")
        self.MoveIndexFF(puls)

    def m2_status(self) -> list:
        buff = []
        buff2 = []
        in1_8 = self.io.input_byte("IN1_8").to_list()
        in9_16 = self.io.input_byte("IN9_16").to_list()

        if in9_16[6] == 1:  # [0,0,0,0,0,0,1,0]:
            m_limit_up = 1
        else:
            m_limit_up = 0
        if in9_16[7] == 1:  # [0,0,0,0,0,0,0,1]:
            m_limit_down = 1
        else:
            m_limit_down = 0

        for i in range(8):
            if i != 0 and in1_8 == [0, 0, 0, 0, 0, 0, 0, 0]:
                buff.append(0)
            if in1_8[i] == 0:
                buff.append(0)
            else:
                buff.append(1)

        for i in range(8):
            if i != 0 and in9_16 == [0, 0, 0, 0, 0, 0, 0, 0]:
                buff2.append(0)
            if in9_16[i] == 0:
                buff2.append(0)
            else:
                buff2.append(1)

        # calculate each digit
        total = (
            buff[0] * 1
            + buff[1] * 2
            + buff[2] * pow(2.0, 2.0)
            + buff[3] * pow(2.0, 3.0)
        ) / 100.0
        total = (
            total
            + (
                buff[4] * 1
                + buff[5] * 2
                + buff[6] * pow(2.0, 2.0)
                + buff[7] * pow(2.0, 3.0)
            )
            / 10.0
        )
        total2 = (
            buff2[0] * 1
            + buff2[1] * 2
            + buff2[2] * pow(2.0, 2.0)
            + buff2[3] * pow(2.0, 3.0)
        )
        total2 = total2 + (buff2[4] * 1) * 10

        m_pos = (total + total2) * pow(-1.0, (buff2[5] + 1))
        return [m_pos, m_limit_up, m_limit_down]

    def um_to_puls(self, dist: int, status: list[int]) -> int:
        self.pulse_rate = self.Config.pulse_rate

        puls = int(dist) * self.pulse_rate
        if (
            dist / 1000.0 + float(status[0]) <= -4.0
            or dist / 1000.0 + float(status[0]) >= 5.5
        ):
            print("move limit")
            return
        if status[1] == 0 and puls < 0:
            print("can't move up direction")
            return
        if status[2] == 0 and puls > 0:
            print("can't move down direction")
            return
        return puls

    def MoveIndexFF(self, puls: int):
        self.cw = self.Config.cw
        self.ccw = self.Config.ccw
        self.motor_speed = self.Config.motor_speed

        if puls >= -65535 and puls <= 65535:
            # index mode
            self.io.output_byte("OUT1_8", [0, 0, 0, 1, 0, 0, 0, 0])
            self.Strobe()
            # step no.
            self.io.output_byte("OUT1_8", [1, 1, 1, 1, 1, 1, 1, 1])
            self.Strobe()
            # position set
            self.io.output_byte("OUT1_8", [0, 0, 0, 0, 0, 0, 1, 1])
            self.Strobe()
            # direction
            if puls >= 0:
                self.io.output_byte("OUT1_8", self.cw, fmt="<I")
                self.Strobe()
            else:
                self.io.output_byte("OUT1_8", self.ccw, fmt="<I")
                self.Strobe()
            # displacement
            self.io.output_byte("OUT1_8", [0, 0, 0, 0, 0, 0, 0, 0])
            self.Strobe()
            self.io.output_byte("OUT1_8", int(abs(puls) / 256), fmt="<I")
            self.Strobe()
            self.io.output_byte("OUT1_8", int(abs(puls) % 256), fmt="<I")
            self.Strobe()
            # start
            self.io.output_byte("OUT1_8", [0, 0, 0, 1, 1, 0, 0, 0])
            self.Strobe()
            time.sleep((abs(puls) / self.motor_speed / 10.0) + 1.0)
            print("Motor stopped")
        else:
            print("Puls number is over.")
            print("Please command x : 10*x [um]")
            return False
        return True

    def Strobe(self):
        time.sleep(0.01)
        self.io.output_byte("OUT9_16", [1, 0, 0, 0, 0, 0, 0, 0])
        time.sleep(0.01)
        self.io.output_byte("OUT9_16", [0, 0, 0, 0, 0, 0, 0, 0])
        time.sleep(0.01)
        return

    # Drive Control
    def drive_move(self, pos) -> None:
        # pos = "on" or "off"
        buff = self.Config.drive_pos[pos.lower()]
        self.io.output_point(buff, 1)
        return

    def contactor_move(self, pos) -> None:
        # pos = "on" or "off"
        buff = self.Config.contactor_pos[pos.lower()]
        self.io.output_point(buff, 9)
        return

    def drive_contactor_status(self) -> list[str, str]:
        pos = self.io.input_byte("IN1_8").to_list()
        if pos[0] == 1 and pos[1] == 1:
            self.contactor_pos = "ON"
        else:
            self.contactor_pos = "OFF"
        if pos[2] == 1 and pos[3] == 1:
            self.drive_pos = "ON"
        else:
            self.drive_pos = "OFF"
        return [self.drive_pos, self.contactor_pos]

    def finalize(self) -> None:
        pass



class Ordinal(int):
    """Integer, with corresponding suffix as ordinal number in string representation.

    Examples
    --------
    >>> Ordinal(1)
    1
    >>> f"{Ordinal(1)}"
    '1st'
    >>> f"{Ordinal(2)} parameter"
    '2nd parameter'

    """

    def __format__(self, format_spec: str) -> str:
        number = str(self)
        if self < 0:
            return number
        elif (self % 10 == 1) and (self % 100 != 11):
            return number + "st"
        elif (self % 10 == 2) and (self % 100 != 12):
            return number + "nd"
        elif (self % 10 == 3) and (self % 100 != 13):
            return number + "rd"
        else:
            return number + "th"



from collections import defaultdict

import astropy.units as u

from .attenuator_base import CurrentAttenuator, NetworkAttenuator


class NetworkAttenuatorSimulator(NetworkAttenuator):
    Manufacturer: str = ""
    Model: str = ""
    Identifier = ""
    is_simulator = True

    def __init__(self) -> None:
        self.loss = defaultdict(lambda: 0)

    def set_loss(self, dB: int, id: str):
        self.loss[id] = dB

    def get_loss(self, id: str) -> u.Quantity:
        return self.loss[id] * u.dB

    def finalize(self) -> None:
        pass

    def close(self) -> None:
        pass


class CurrentAttenuatorSimulator(CurrentAttenuator):
    Manufacturer: str = ""
    Model: str = ""
    Identifier = ""
    is_simulator = True

    def __init__(self) -> None: ...

    def get_outputrange(self, id: int) -> dict: ...

    def set_outputrange(self, id: int, outputrange: str): ...

    def output_current(self, id: int, current: float): ...

    def finalize(self): ...

    def close(self): ...



import time

import astropy.units as u
import ogameasure

from ... import get_logger
from ...core.security import busy
from .thermometer_base import Thermometer


class Model218(Thermometer):
    """Thermometer, which can check tempareture of cryostat.

    Notes
    -----

    Configuration items for this device:

    communicator : str
        Communicator of thermometer. GPIB or USB can be chosen.

    host : str
        IP address for GPIB communicator.
        If you use GPIB communicator, you must set this parameter.

    port : int
        GPIB port of using devices. Please check device setting.
        If you use GPIB communicator, you must set this parameter.

    usb_port : str
        USB port of using devices.
        If you use USB communicator, you must set this parameter.

    channel : Dict[str, int]
        Human-readable channel name. The value should be
        mapping from human readableversion (str) to
        device level identifier (int). You can assign any name to the
        channels.

    """

    Manufacturer = "LakeShore"
    Model = "Model218"

    Identifier = "communicator"

    def __init__(self) -> None:
        self.logger = get_logger(self.__class__.__name__)

        if self.Config.communicator == "GPIB":
            com = ogameasure.gpib_prologix(
                host=self.Config.host, gpibport=self.Config.port
            )
            self.thermometer = ogameasure.Lakeshore.model218(com)
        elif self.Config.communicator == "USB":
            self.thermometer = ogameasure.Lakeshore.model218_usb(self.Config.usb_port)
        else:
            self.logger.warning(
                f"There is not exsited communicator: {self.Config.communicator}."
                "Please choose USB or GPIB."
            )
        com.send("KRDG? 0")
        time.sleep(0.1)
        _ = com.readline()

    def get_temp(self, id: str) -> u.Quantity:
        ch = self.Config.channel[id]
        with busy(self, "busy"):
            data = self.thermometer.kelvin_reading_query(ch)
            time.sleep(0.1)
            return data * u.K

    def finalize(self) -> None:
        self.thermometer.com.close()

    def close(self) -> None:
        self.thermometer.com.close()



from abc import abstractmethod

import astropy.units as u

from ..device_base import DeviceBase


class NetworkAttenuator(DeviceBase):
    @abstractmethod
    def set_loss(self, dB: int, id: str): ...

    @abstractmethod
    def get_loss(self, id: str) -> u.Quantity: ...

    @abstractmethod
    def close(self) -> None: ...


class CurrentAttenuator(DeviceBase):
    @abstractmethod
    def get_outputrange(self, ch: int) -> dict: ...

    @abstractmethod
    def set_outputrange(self, id: int, outputrange: str): ...

    @abstractmethod
    def set_current(self, id: str, mA: float): ...

    @abstractmethod
    def apply_current(self): ...

    @abstractmethod
    def finalize(self): ...

    @abstractmethod
    def close(self): ...



from abc import abstractmethod

import astropy.units as u

from ..device_base import DeviceBase


class WeatherStation(DeviceBase):

    Manufacturer: str = ""
    Model: str

    @abstractmethod
    def get_temperature(self) -> u.Quantity: ...

    @abstractmethod
    def get_humidity(self) -> float: ...

    @abstractmethod
    def get_pressure(self) -> u.Quantity: ...

    @abstractmethod
    def finalize(self) -> None: ...

    @abstractmethod
    def close(self) -> None: ...



import astropy.units as u

from ...core.math import Random
from ...core.units import dBm
from .power_meter_base import PowerMeter


class PowerMeterSimulator(PowerMeter):
    Manufacturer: str = ""
    Model: str = ""
    Identifier = ""
    is_simulator = True

    def __init__(self) -> None:
        self._rand = Random().walk(-16, 0.1, -1)

    def get_power(self) -> u.Quantity:
        return next(self._rand) * dBm

    def finalize(self) -> None:
        pass

    def close(self) -> None:
        pass



from abc import abstractmethod
from typing import Union

import astropy.units as u

from ..device_base import DeviceBase


class SignalGenerator(DeviceBase):

    Manufacturer: str = ""
    Model: str

    @abstractmethod
    def set_freq(self, GHz: Union[int, float]) -> None: ...

    @abstractmethod
    def set_power(self, dBm: Union[int, float]) -> None: ...

    @abstractmethod
    def get_freq(self) -> u.Quantity: ...

    @abstractmethod
    def get_power(self) -> u.Quantity: ...

    @abstractmethod
    def start_output(self) -> None: ...

    @abstractmethod
    def stop_output(self) -> None: ...

    @abstractmethod
    def get_output_status(self) -> bool: ...

    @abstractmethod
    def finalize(self) -> None: ...

    @abstractmethod
    def close(self) -> None: ...



import astropy.units as u
import ogameasure

from ... import get_logger
from ...core.security import busy
from ...utils import skip_on_simulator
from .attenuator_base import NetworkAttenuator


class A11713C(NetworkAttenuator):
    """Attenuator, which can attennuate four IF sigal power.

    Notes
    -----

    Configuration items for this device:

    communicator : str
        Communicator of thermometer. GPIB or LAN can be chosen.

    host : str
        IP address for GPIB and ethernet communicator.

    gpib_port : int
        GPIB port of using devices. Please check device setting.
        If you use GPIB communicator, you must set this parameter.

    lan_port : int
        LAN port of using devices. This parameter is setted to 5025 by manufacturer.
        If you use LAN communicator, you must set this parameter.

    model: str
        Attenuator model which you use in this device.
        Following model are available:
        "NA", "AG8494g", "AG8494h", "AG8495g", "AG8495h", "AG8495k",
        "AG8496g", "AG8496h", "AG8497k", "AG84904k", "AG84904l",
        "AG84904m", "AG84905m", "AG84906k", "AG84906l", "AG84907k",
        "AG84907l", "AG84908m"
        For example:
        `{ 1LU = "AG8494g", 1LL = "AG8495k", 1RU = "AG84905m", 1RL = "AG84907k"}`

    channel : Dict[str]
        Human-readable channel name. The value should be
        mapping from human readableversion (str) to
        device level identifier (int). You can assign any name to the
        channels up to four channels: "1X" (BANK1, X), "1Y" (BANK1, Y),
        "2X" (BANK2, X), "2Y" (BANK2 Y).
        For example: `{ 1LU = "1X", 1LL = "1Y", 1RU = "2X", 1RL = "2Y"}`

    """

    Manufacturer = "Agilent"
    Model = "11713C"

    Identifier = "host"

    @skip_on_simulator
    def __init__(self) -> None:
        self.logger = get_logger(self.__class__.__name__)

        if self.Config.communicator == "GPIB":
            com = ogameasure.gpib_prologix(self.Config.host, self.Config.gpib_port)
        elif self.Config.communicator == "LAN":
            com = ogameasure.ethernet(self.Config.host, self.Config.lan_port)
        else:
            self.logger.warning(
                f"There is not exsited communicator: {self.Config.communicator}."
                "Please choose USB or GPIB."
            )
        self.io = ogameasure.Agilent.agilent_11713C(com)
        self.model_check()
        pass

    def model_check(self) -> None:
        for id in self.Config.model.keys():
            bank = int(self.Config.channel[id][0])
            ch = self.Config.channel[id][1]
            model = self.Config.model[id]
            dev_model = self.io.att_model_query(ch, bank)
            if model != "AG8495k":
                _model = model[:-1]
            else:
                _model = model[:-1] + model[-1].upper()
            if _model != dev_model:
                raise ValueError(
                    "Attenutor model in config is not match with"
                    f"the model which you set in device. {id}: {bank}, {ch}, {model}"
                )
            pass

    def get_loss(self, id: str) -> u.Quantity:
        with busy(self, "busy"):
            bank = int(self.Config.channel[id][0])
            ch = self.Config.channel[id][1]
            try:
                return self.io.att_level_query(ch, bank) * u.dB
            except IndexError:
                pass
            raise ValueError(f"Invalid channel: {ch}")

    def set_loss(self, dB: int, id: str) -> None:
        with busy(self, "busy"):
            bank = int(self.Config.channel[id][0])
            ch = self.Config.channel[id][1]
            self.io.att_level_set(dB, ch, bank)

    def finalize(self) -> None:
        self.io.com.close()

    def close(self) -> None:
        self.finalize()



import astropy.units as u

from .motor_base import Motor


class MotorSimulator(Motor):
    Manufacturer: str = ""
    Model: str = ""
    Identifier = ""
    is_simulator = True

    def __init__(self) -> None:
        self.motion = ...

    def set_step(self, step: int, axis: str) -> None:
        """Drive to (maybe device-specific) absolute position."""
        raise NotImplementedError

    def set_speed(self, speed: float, axis: str) -> None:
        raise NotImplementedError

    def get_step(self, axis: str) -> int:
        """Maybe device-specific absolute position."""
        raise NotImplementedError

    def get_speed(self, axis: str) -> u.Quantity:
        raise NotImplementedError

    def finalize(self) -> None:
        pass



from itertools import count
from typing import Generator

import astropy.units as u

from .observation_spec_base import (
    ObservationMode,
    ObservationSpec,
    TimeKeeper,
    Waypoint,
)


class RadioPointingSpec(ObservationSpec):
    __slots__ = ("_hot_time_keeper", "_off_time_keeper")
    _repr_frame = "altaz"

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        # Cannot define this in scan mode, so use default value of 1 in such case.
        _points_per_scan = max(1, (self["method"] + 1) / 2)

        self._hot_time_keeper = TimeKeeper(self["load_interval"], _points_per_scan)
        self._off_time_keeper = TimeKeeper(self["off_interval"], _points_per_scan)

    def observe(self) -> Generator[Waypoint, None, None]:
        iterate_counter = count() if self["n"] < 0 else range(self["n"])
        _method = self["method"]

        for iteration_count in iterate_counter:
            coords = self._points() if _method > 0 else self._scan()
            unit = "point" if _method > 0 else "scan"

            for i, coord in enumerate(coords):
                if self._hot_time_keeper.should_observe:
                    self._hot_time_keeper.tell_observed()
                    yield self.hot(f"{iteration_count}-{i}")

                if self._off_time_keeper.should_observe:
                    self._off_time_keeper.tell_observed()
                    yield self.off(f"{iteration_count}-{i}")

                self._hot_time_keeper.increment(unit)
                self._off_time_keeper.increment(unit)
                coord.id = f"{iteration_count}-{i}"
                yield coord

        # Should be executed at the end of the observation, to enable interpolation.
        yield self.hot("9999")
        yield self.off("9999")

    @property
    def _on_point(self) -> Waypoint:
        return Waypoint(
            mode=ObservationMode.ON,
            reference=self._reference,
            integration=self["integ_on"],
            offset=(self["offset_az"], self["offset_el"], "altaz"),
        )

    def _points(self) -> Generator[Waypoint, None, None]:
        n = self["n"]
        if (n > 0) and (n % 4 != 1):
            raise ValueError("n must be positive and 1 + 4 * N.")

        on_point = self._on_point.reference

        n_points_per_arm = int((self["method"] - 1) / 4)
        for idx in range(-1 * n_points_per_arm, n_points_per_arm + 1):
            offset = (
                self["offset_az"] + self["grid_az"] * idx,
                self["offset_el"],
                "altaz",
            )
            yield Waypoint(
                mode=ObservationMode.ON,
                reference=on_point,
                offset=offset,
                integration=self["integ_on"],
            )
        for idx in range(-1 * n_points_per_arm, n_points_per_arm + 1):
            offset = (
                self["offset_az"],
                self["offset_el"] + self["grid_el"] * idx,
                "altaz",
            )
            yield Waypoint(
                mode=ObservationMode.ON,
                reference=on_point,
                offset=offset,
                integration=self["integ_on"],
            )

    def _scan(self) -> Generator[Waypoint, None, None]:
        yield Waypoint(
            mode=ObservationMode.ON,
            start=(-1 * self["max_separation_az"], 0 << u.deg),
            stop=(self["max_separation_az"], 0 << u.deg),
            reference=self._reference,
            scan_frame="altaz",
            speed=self["speed"],
            offset=(self["offset_az"], self["offset_el"], "altaz"),
        )
        yield Waypoint(
            mode=ObservationMode.ON,
            start=(0 << u.deg, self["max_separation_el"]),
            stop=(0 << u.deg, -1 * self["max_separation_el"]),
            reference=self._reference,
            scan_frame="altaz",
            speed=self["speed"],
            offset=(self["offset_az"], self["offset_el"], "altaz"),
        )



from .convert import *  # noqa: F401, F403
from .frame import describe_frame, parse_frame  # noqa: F401
from .observations import *  # noqa: F401, F403
from .observer import Observer  # noqa: F401
from .optimize import DriveLimitChecker  # noqa: F401
from .path_finder import CoordinateGeneratorManager, PathFinder  # noqa: F401
from .pointing_error import PointingError  # noqa: F401



from typing import Callable, Dict, List

import astropy.units as u

from ...core.security import busy, sanitize
from .ad_converter_base import ADConverter


class CPZ3177(ADConverter):
    """a/d converter, which can convert by 32 or 64 channels.

    Notes
    -----
    Configuration items for this device:

    rsw_id : {0, 1, ..., 16} or {"0", "1", ..., "9", "A", ..., "F"}
        Board identifier. This should be set to the same value as the rotary switch
        "RSW1" mounted on the side of the board. The board is shipped with default RSW1
        setting of 0. This ID would be non-zero, when multiple PCI board of same model
        are mounted on a single FA (Factory Automation) controller.

    ave_num : int
        Number of sampled data to be averaged over. This script not only measures the
        actual voltage, but also performs fluctuation reduction in order to measure a
        faint voltage. The value should not be too large compared to sampling frequency
        ``smpl_freq``, otherwise the output won't follow the change of voltage.

    smpl_freq : int
        Sampling frequency; number of data obtained in 1 second. The value should not
        be too small against ``ave_num``.

    single_diff : {"SINGLE" or "DIFF"}
        Input type of voltage. "SINGLE" is "Single-ended input", "DIFF" is
        "Differential input".

    all_ch_num : int
        Number of channels used in voltage measurement. The maximum number is
        32 in differential input, 64 in single-ended input. Please read the manual
        of this board for wiring design. It must be defined the maximum channel number
        to be used. For example, when channels [3, 5, 9, 12] are used, ``all_ch_mum``
        must be set ``12`` at least.

    ch_range : {'0_5V', '0_10V', '2P5V', '5V' or '10V'}
        Voltage measurement range. The acceptable "range" values are as below:
        '0_5V': 0 - 5 V,
        '0_10V' : 0 - 10 V,
        '2P5V' : -2.5 - 2.5 V,
        '5V' : -5 - 5 V,
        '10V' : -10 - 10 V.
        These should be set to the same value as the value corresponding to combination
        of three DIP switch "DSW1", "DSW2", "DSW3" mounted on the side of the board.
        Please read the manual of this board for the combination of DIP switches.

    channel : Dict[str, int]
        Human-readable channel name. The value should be mapping from human readable
        version (str) to device level identifier (int). You can assign any name to the
        channels, but you must name usage at the head of channel name; "sis" or "hemt",
        like "sis_USB_V" or "hemt_ch1_Vdr". No need to define the aliases for
        all the channels listed in ``smpl_ch_req``, but defining aliases for
        unused channels will raise error.

    converter : List[Dict[str, str]]
        Functions to convert measured voltage to any parameter you want, in format
        ``{ch = str(channel id defined in ``channel``),
        str(parameter_type) = str(function), units = str(units of read value)}``.
        Supported``parameter_types`` are ["V", "I"], and ``x`` in ``function``
        will be substituted by the measured value. This would be useful when measured
        voltage is scaled and/or shifted version of physical parameter.

    You can use this PCI board for some target, like SIS and HEMT bias
    voltage measurement. You must wrap up all setting in one representative target,
    rest targets have only device model name; ``_`` and ``rsw_id``.
    Different measurement setting in ``ave_num``, ''smpl_freq``, ``single_diff``,
    ``all_ch_num`` and  ``ch_range`` between different target will raise error.
    See defaults setting file in ``neclib/defaults/config.toml``.

    """

    Manufacturer = "Interface"
    Model = "CPZ3177"

    Identifier = "rsw_id"

    def __init__(self) -> None:
        import pyinterface

        self.rsw_id = self.Config.rsw_id
        self.ave_num = self.Config.ave_num
        self.smpl_freq = self.Config.smpl_freq
        self.single_diff = self.Config.single_diff
        self.all_ch_num = self.Config.all_ch_num
        self.ch_range = self.Config.ch_range
        self.smpl_ch_req = [
            {"ch_no": i, "range": self.ch_range}
            for i in range(1, self.all_ch_num + 1, 1)
        ]

        self.ad = pyinterface.open(3177, self.rsw_id)
        self.ad.stop_sampling()
        self.ad.initialize()
        self.ad.set_sampling_config(
            smpl_ch_req=self.smpl_ch_req,
            smpl_num=1000,
            smpl_freq=self.smpl_freq,
            single_diff=self.single_diff,
            trig_mode="ETERNITY",
        )
        self.ad.start_sampling("ASYNC")

    def get_data(self) -> List[float]:
        with busy(self, "busy"):
            offset = self.ad.get_status()["smpl_count"] - self.ave_num
            data = self.ad.read_sampling_buffer(self.ave_num, offset)
            data_li_2 = []
            for i in range(self.all_ch_num):
                data_li = []
                for k in range(self.ave_num):
                    data_li.append(data[k][i])
                data_li_2.append(data_li)

            ave_data_li = []
            for data in data_li_2:
                d = sum(data) / self.ave_num
                ave_data_li.append(d)
            return ave_data_li

    @property
    def converter(self) -> Dict[str, Callable[[float], float]]:
        conv = []
        for i in self.Config.converter:
            _ = [sanitize(expr, "x") for k, expr in i.items() if k == "func"]
            conv.append(
                {k: v if k != "func" else eval(f"lambda x: {v}") for k, v in i.items()}
            )
        return conv

    def get_all(self, target: str) -> dict:
        data = self.get_data()
        data_dict = {}
        filtered_conv = list(
            filter(lambda item: item["ch"].startswith(target), self.converter)
        )
        for i in filtered_conv:
            ch = self.Config.channel[i["ch"]]
            value = i["func"](data[ch - 1])
            data_dict[i["ch"]] = u.Quantity(value, i["units"])
        return data_dict

    def get_from_id(self, id: str) -> u.Quantity:
        ch = self.Config.channel[id]
        li_search = list(filter(lambda item: item["ch"] == id, self.converter))[0]
        value = li_search["func"](self.get_data()[ch - 1])
        return u.Quantity(value, li_search["units"])

    def finalize(self) -> None:
        self.ad.stop_sampling()

    def close(self) -> None:
        self.finalize()



from .e8257d import E8257D  # noqa: F401
from .fsw0010 import FSW0010  # noqa: F401
from .fsw0020 import FSW0020  # noqa: F401
from .mg3692c import MG3692C  # noqa: F401
from .simulator import SignalGeneratorSimulator  # noqa: F401



from __future__ import annotations

from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass, fields
from typing import TYPE_CHECKING, Any, Dict, Generator, Optional, Tuple, Union

import astropy.units as u
from astropy.coordinates import BaseCoordinateFrame

from ...core.normalization import NPArrayValidator
from ...core.types import CoordFrameType, DimensionLess, UnitType

if TYPE_CHECKING:
    from ..convert import CoordCalculator, Coordinate

T = Union[DimensionLess, u.Quantity]


class Path(ABC):

    tight: bool
    infinite: bool
    waypoint: bool

    def __init__(
        self, calc: CoordCalculator, *target, unit: Optional[UnitType] = None, **kwargs
    ) -> None:
        self._calc = calc

        if len(target) == 0:
            _target = None
        elif (len(target) == 1) and isinstance(target[0], str):
            _target = calc.name_coordinate(target[0])
        elif len(target) == 3:
            lon, lat, frame = target
            _target = calc.coordinate(lon=lon, lat=lat, frame=frame, unit=unit)
        else:
            raise TypeError(
                "Invalid number of positional arguments: expected 0 (none, other "
                "arguments specify the absolute coordinate), 1 (target_name) or "
                f"3 (lon, lat, coordinate_frame), but got {len(target)}"
            )
        self._target: Optional[Coordinate] = _target

    @property
    @abstractmethod
    def arguments(self) -> Tuple[Tuple[Any, ...], Dict[str, Any]]: ...

    @property
    def target_frame(self) -> Optional[CoordFrameType]:
        if self._offset is not None:
            return self._offset.frame

        if self._target is None:
            return None
        else:
            import time

            target = (
                self._target.realize(time=time.time())  # type: ignore
                if hasattr(self._target, "realize")
                else self._target
            )
            frame = target.frame

        if isinstance(frame, BaseCoordinateFrame):
            return frame.replicate_without_data()
        else:
            return frame


@dataclass
class Index:
    time: NPArrayValidator = NPArrayValidator[float]()
    index: NPArrayValidator = NPArrayValidator[Union[int, float]]()


@dataclass
class ControlContext:
    tight: Optional[bool] = None
    """If True, the section is for observation so its accuracy should be inspected."""
    start: Optional[float] = None
    """Start time of this control section."""
    stop: Optional[float] = None
    """End time of this control section."""
    duration: Optional[float] = None
    """Time duration of this control section."""
    infinite: bool = False
    """Whether the control section is infinite hence need interruption or not."""
    waypoint: bool = False
    """Whether this is waypoint hence need some value to be sent or not."""

    @contextmanager
    def properties_modified(self, **kwargs) -> Generator[None, None, None]:
        original = {
            k: getattr(self, k) for k in kwargs if k in self.__dataclass_fields__
        }
        for k, v in kwargs.items():
            setattr(self, k, v)
        try:
            yield
        finally:
            for k, v in original.items():
                setattr(self, k, v)

    def update(self, other: ControlContext, **kwargs: Any) -> None:
        for k in fields(self):
            v = kwargs.get(k.name, getattr(other, k.name))
            setattr(self, k.name, v)



from __future__ import annotations

import warnings
from collections import defaultdict
from dataclasses import fields
from io import StringIO
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt

from ..exceptions import NECSTAccessibilityWarning

if TYPE_CHECKING:
    from ..data_type.status_manager import StatusManager


def html_repr_of_status(a: StatusManager):
    colors = defaultdict(lambda: f"C{len(colors)}")
    data_fields = [f for f in fields(a.ctx_type) if f.name not in [a.start, a.stop]]
    max_field_name_length = max(len(f.name) for f in data_fields)

    with plt.style.context("Solarize_Light2"), plt.rc_context(  # type: ignore
        {"font.family": "monospace", "font.size": 9}
    ):
        height = len(data_fields) * 0.5
        width = max_field_name_length * 0.05 + 5
        fig, ax = plt.subplots(figsize=(width, height))
        _x = {_a.start for _a in a.ctx} | {_a.stop for _a in a.ctx}
        x = sorted(_x)
        d = [a.get(_x) for _x in x]

        for _x, _d in zip(x, d):
            start, stop = getattr(_d, a.start), getattr(_d, a.stop)
            if (start is None) or (stop is None):
                continue
            width = stop - start

            for i, field in enumerate(data_fields):
                name = field.name
                id = str(getattr(_d, name))
                ax.barh(i, width, left=start, height=0.8, color=colors[id])
                ax.text(
                    start + width / 2, i, getattr(_d, name), ha="center", va="center"
                )
        ax.set(
            xlim=(x[0], x[-1]),
            yticks=range(len(data_fields)),
            yticklabels=[f.name for f in data_fields],
        )
        ax.grid(False)
        ax.invert_yaxis()

        warnings.simplefilter("error", UserWarning)
        try:
            fig.tight_layout()
        except UserWarning:
            warnings.simplefilter("default", UserWarning)
            warnings.warn(
                "tight_layout failed. This can be caused by non-zero default value of "
                "`start` and/or `stop` attributes.",
                NECSTAccessibilityWarning,
            )

        _svg = StringIO()
        fig.savefig(_svg, format="svg")
        plt.close()
        _svg.seek(0)
        return _svg.read()



from typing import Dict, Type, Union

from .. import config, get_logger, utils
from ..core.exceptions import NECSTConfigurationError
from .device_base import DeviceBase, Devices

logger = get_logger(__name__)


def get_device_list() -> Dict[str, Type["DeviceBase"]]:
    return {d.Model: d for d in DeviceBase._implementations}


def get_device_configuration() -> Dict[str, Union[str, Dict[str, str]]]:
    devices = {k[:-2]: v for k, v in config.items() if k.endswith("._")}
    device_kinds = set(d.split(".")[0] for d in devices)
    parsed = {}
    for kind in device_kinds:
        if kind in devices:
            parsed[kind] = devices[kind]
        else:
            prefix_length = len(f"{kind}.")
            parsed[kind] = {
                k[prefix_length:]: v for k, v in devices.items() if k.startswith(kind)
            }
    return parsed


def parse_device_configuration() -> Dict[str, Devices]:
    configuration = get_device_configuration()
    implementations = {}
    for k, v in configuration.items():
        try:
            impl = DeviceBase.bind(name=k, model=v)
            implementations[k] = impl if isinstance(impl, Devices) else Devices(impl)
            implementations[utils.toCamelCase(k)] = implementations[k]
        except Exception:
            raise NECSTConfigurationError(
                f"Driver implementation for device '{v}' ({k}) not found."
            )
    return implementations


def list_implementations() -> Dict[str, Type[DeviceBase]]:
    return get_device_list()



from typing import Any, Callable, List, Optional, TypeVar, Union, overload

import astropy.units as u
import numpy as np
import numpy.typing as npt

from ..types import DimensionLess

T = TypeVar("T", bound=Union[DimensionLess, npt.NDArray[Any], u.Quantity])


@overload
def frange(
    stop: Union[DimensionLess, npt.NDArray[Any], u.Quantity],
    /,
    *,
    inclusive: bool = False,
    pad_method: Union[str, Callable[..., Any]] = "edge",
    **pad_kwargs: Any,
) -> npt.NDArray[Any]: ...


@overload
def frange(
    start: T,
    stop: T,
    step: Optional[T] = None,
    /,
    *,
    inclusive: bool = False,
    pad_method: Union[str, Callable[..., Any]] = "edge",
    **pad_kwargs: Any,
) -> npt.NDArray[Any]: ...


def frange(
    a: T,
    b: Optional[T] = None,
    c: Optional[T] = None,
    /,
    *,
    inclusive: bool = False,
    pad_method: Union[str, Callable[..., Any]] = "edge",
    **pad_kwargs: Any,
) -> npt.NDArray[Any]:
    """A range function with support for quantities, float values, and arrays.

    Parameters
    ----------
    start
        First value to be generated.
    stop
        Last value to be generated is equal to or less than this value.
    step
        Spacing between values.
    inclusive
        If ``True``, ``stop`` is included in the range, no matter what ``step`` is.
    pad_method
        Method to use for padding the last value to match the length of the other
        values. See
        `numpy.pad <https://numpy.org/doc/stable/reference/generated/numpy.pad.html>`_
        for more details.
    pad_kwargs
        Keyword arguments for
        `numpy.pad <https://numpy.org/doc/stable/reference/generated/numpy.pad.html>`_.

    Returns
    -------
    Array of values.

    Raises
    ------
    TypeError
        If arguments are mixture of quantities or non-quantities.
    UnitConversionError
        If any of the arguments have non-equivalent physical dimension.
    UnitConversionError
        If any of the arguments are given as sequence of quantities.

    Examples
    --------
    >>> frange(1, 10, 2)
    array([1, 3, 5, 7, 9, 10])
    >>> frange(1 * u.m, 10 * u.m, 2 * u.m)
    <Quantity [ 1.,  3.,  5.,  7.,  9., 10.] m>
    >>> frange([1, 2], [2, 2.1], [3, 1])
    array([[1. , 2. ],
           [2. , 2.1]])

    """
    if b is None:
        return frange(np.zeros_like(a), a, np.ones_like(a))
    if c is None:
        return frange(a, b, np.ones_like(a))  # type: ignore

    # Splitting the unit is workaround for the issue:
    # https://docs.astropy.org/en/stable/known_issues.html#numpy-array-creation-functions-cannot-be-used-to-initialize-quantity
    unit: Union[int, u.UnitBase] = 1
    if (
        isinstance(a, u.Quantity)
        and isinstance(b, u.Quantity)
        and isinstance(c, u.Quantity)
    ):
        unit = a.unit  # type: ignore
        a, b, c = a.to_value(unit), b.to_value(unit), c.to_value(unit)  # type: ignore
    if (
        isinstance(a, u.Quantity)
        or isinstance(b, u.Quantity)
        or isinstance(c, u.Quantity)
    ):
        raise TypeError(
            "All arguments must be quantities or non-quantities, not mixture of them."
        )

    start, stop, step = np.asanyarray(a), np.asanyarray(b), np.asanyarray(c)
    start, stop, step = np.broadcast_arrays(start, stop, step)
    zero_dimensional: bool = start.ndim == 0

    sequence: List[npt.NDArray[Any]] = [
        np.arange(_start, _stop, _step)
        for _start, _stop, _step in zip(start.flat, stop.flat, step.flat)
    ]

    # Ensure stop value is covered
    if inclusive:
        sequence = [
            np.pad(s, (0, 1), "constant", constant_values=_stop)
            for s, _stop in zip(sequence, stop.flat)
        ]
    length = max(len(s) for s in sequence)
    sequence = [
        np.pad(s, (0, length - len(s)), pad_method, **pad_kwargs)  # type: ignore
        for s in sequence
    ]
    ret = np.asanyarray(sequence[0] if zero_dimensional else sequence)

    if inclusive:
        stop_is_duplicated = np.bool_(ret[..., -2] == ret[..., -1])
    else:
        stop_is_duplicated = np.bool_(False)

    if isinstance(stop_is_duplicated, np.ndarray):
        if stop_is_duplicated.size == 0:
            stop_is_duplicated = np.bool_(False)
        stop_is_duplicated = stop_is_duplicated.all()

    ret = ret * unit
    if stop_is_duplicated:
        ret = ret[..., :-1]
    return ret if zero_dimensional else np.reshape(ret, (*start.shape, -1))



from typing import Any, Generic, TypeVar

import numpy as np
import numpy.typing as npt

T = TypeVar("T")


class NPArrayValidator(Generic[T]):
    """Type validator, to force some instance variables to be Quantity."""

    # For details on validator descriptor design, see:
    # https://docs.python.org/3/howto/descriptor.html#validator-class

    def __set_name__(self, owner: Any, name: str) -> None:
        self.private_name = "_" + name

    def __get__(self, instance: Any, owner: Any) -> npt.NDArray[T]:  # type: ignore
        if instance is None:
            raise AttributeError(f"Cannot access attribute {self.private_name!r}")
        return getattr(instance, self.private_name)

    def __set__(self, instance: Any, value: npt.ArrayLike) -> None:
        if not isinstance(value, np.ndarray):
            value = np.asanyarray(value)
        setattr(instance, self.private_name, value)



from typing import Generator, Optional, Sequence, Tuple, Union, overload

import numpy as np

from . import Functions, clip


class Random:
    def __init__(self, seed: Optional[int] = None, limits: Optional[Sequence] = None):
        self.seed = seed
        self.rng = np.random.default_rng(seed)
        self.limits = limits

    def _walk(
        self,
        x: Union[Sequence, float, int, np.ndarray, None],
        mu: Union[int, float],
        noise: Union[int, float],
        stability: Union[int, float],
    ) -> Union[np.ndarray, float]:
        x = np.asanyarray(x) if x is not None else mu
        fluctuation = self.rng.uniform(-noise / 2, noise / 2)
        stability = Functions.Sigmoid(0, 1)(stability)
        f2 = self.rng.uniform(0, stability)
        drift = self.rng.normal(f2 * (mu - x), noise)
        drift_applied = mu if x is None else x + drift

        if self.limits is None:
            return drift_applied + fluctuation
        return clip(drift_applied + fluctuation, *self.limits)  # type: ignore

    @overload
    def walk(
        self,
        mu: Union[int, float],
        noise: Union[int, float],
        stability: Union[int, float],
        *,
        initial: Union[int, float, None] = None,
    ) -> Generator[float, None, None]: ...

    @overload
    def walk(
        self,
        mu: Union[int, float],
        noise: Union[int, float],
        stability: Union[int, float],
        *,
        initial: Union[Sequence, np.ndarray],
    ) -> Generator[np.ndarray, None, None]: ...

    def walk(
        self,
        mu: Union[int, float],
        noise: Union[int, float],
        stability: Union[int, float],
        *,
        initial: Union[Sequence, float, int, np.ndarray, None] = None,
    ) -> Generator[Union[np.ndarray, float], None, None]:
        x = initial
        while True:
            x = self._walk(x, mu, noise, stability)
            yield x

    @overload
    def fluctuation(
        self,
        noise: Union[int, float],
        *,
        shape: None = None,
    ) -> Generator[float, None, None]: ...

    @overload
    def fluctuation(
        self,
        noise: Union[int, float],
        *,
        shape: Tuple[int, ...],
    ) -> Generator[np.ndarray, None, None]: ...

    def fluctuation(
        self,
        noise: Union[int, float],
        *,
        shape: Optional[Tuple[int, ...]] = None,
    ) -> Generator[Union[np.ndarray, float], None, None]:
        shape = shape if shape is not None else ()
        while True:
            yield self.rng.uniform(-noise / 2, noise / 2, shape)



"""TOML file handling."""

import os
from typing import IO, Any, Dict, Mapping, Union

from tomlkit import TOMLDocument, dumps, parse
from tomlkit.items import Table

from .general import read as read_file


def read(__file: Union[os.PathLike, str, IO], /) -> TOMLDocument:
    r"""Read and parse a TOML file.

    Parameters
    ----------
    __file
        Path to the TOML file or a file-like object.

    Returns
    -------
    parsed
        The parsed content.

    Examples
    --------
    >>> neclib.core.toml.read("path/to/pyproject.toml")

    You can also pass a file-like object:

    >>> with open("pyproject.toml") as f:
    ...     neclib.core.toml.read(f)

    """
    return parse(read_file(__file))


def flatten(
    doc: Dict[str, Any], /, *, prefix: str = "", sep: str = "."
) -> TOMLDocument:
    """Flatten a parsed TOML object.

    Parameters
    ----------
    doc
        TOML document or TOML Table to be flattened.
    prefix
        Prefix to be added to the keys.
    sep
        Separator to be used to join keys.

    Returns
    -------
    flattened
        The flattened TOML document.

    Notes
    -----
    This function flattens dict-like objects recursively, but ``InlineTable`` objects
    will be kept as is. This feature comes from the following motivations:

    - TOML files should be well-structured by the parameters semantics.
    - Deeply nested dict-like object is hard to handle, so we flatten them.
    - Flattening all dict-like objects will remove ability to store
      structured-parameter (dict-like object) in TOML files.

    The author never felt this implementation a good one, since it contradicts the TOML
    specification: "inline-tables are just a short-hand definition of tables", which
    implies they should be treated equivalently. If the alternative method to flatten
    TOML document is found, this function will employ it.

    Examples
    --------
    >>> neclib.core.toml.flatten({"a": {"b": 1, "c": 2}, "d": 3})
    {"a.b": 1, "a.c": 2, "d": 3}

    """
    ret = TOMLDocument()
    prefix = prefix + sep if prefix else ""
    for k, v in doc.items():
        if isinstance(v, Table):
            ret.update(flatten(v, prefix=prefix + k))
        else:
            ret.update({prefix + k: v})
    return ret


def to_string(__mapping: Mapping[str, Any], /) -> str:
    r"""Convert a mapping to a TOML string.

    Parameters
    ----------
    __mapping
        Mapping object to be converted.

    Returns
    -------
    toml_string
        TOML format representation of given mapping object.

    Examples
    --------
    >>> neclib.core.toml.to_string({"a": 1, "b": 2})
    'a = 1\nb = 2'

    """
    return dumps(__mapping)



from .a11713b import A11713B  # noqa: F401
from .a11713c import A11713C  # noqa: F401
from .cpz340516 import CPZ340516  # noqa: F401
from .rhio10 import RHIO10  # noqa: F401
from .simulator import CurrentAttenuator  # noqa: F401
from .simulator import NetworkAttenuator  # noqa: F401



import time
from typing import Dict, Optional, Tuple

import astropy.units as u
import psutil
from astropy.units import Quantity


class LoadChecker:
    cpu_count = psutil.cpu_count()
    """Number of logical processors.

    Number of threads that can be handled simultaneously. Normally this is the same as
    the number of CPU cores. If simultaneous multi-threading is enabled (e.g., Intel's
    Hyper-Threading technology), this value would be larger than the number of CPU
    cores.

    """

    def __init__(self):
        self._last_netcount: Optional[Dict[str, float]] = None

        # Ensure that the first call of methods returns meaningful values.
        self.cpu_usage()
        self.network_amount()

    def cpu_usage(self) -> Quantity:
        """Mean CPU usage since last call.

        This method averages the usage from last call or class initialization.

        """
        usage = psutil.cpu_percent(interval=None, percpu=True)
        return usage * u.percent

    def loadavg(self) -> Tuple[float, float, float]:
        """Load average per processor.

        Returns
        -------
        Number of processes using or waiting to use the processor, averaged over all
        logical processors. The load is averaged over time duration of 1min, 5min, and
        15min.

        """
        return tuple(load / self.cpu_count for load in psutil.getloadavg())

    def memory_available(self) -> Quantity:
        """Available memory."""
        return psutil.virtual_memory().available * u.byte

    def disk_usage(self) -> Quantity:
        """Disk (storage) usage."""
        return psutil.disk_usage("/").percent * u.percent

    def network_amount(self) -> Dict[str, Quantity]:
        """Network communication amount.

        This method averages the data rate from last call or class initialization.

        """
        netcount = psutil.net_io_counters()
        current_netcount: Dict[str, float] = {
            "sent": netcount.bytes_sent,
            "recv": netcount.bytes_recv,
            "time": time.monotonic(),
        }
        if self._last_netcount is None:
            self._last_netcount = current_netcount
            return {"sent": 0 * u.byte / u.s, "recv": 0 * u.byte / u.s}

        dt = current_netcount["time"] - self._last_netcount["time"]
        dsent = netcount.bytes_sent - self._last_netcount["sent"]
        drecv = netcount.bytes_recv - self._last_netcount["recv"]
        self._last_netcount = current_netcount
        return {"sent": dsent / dt * u.byte / u.s, "recv": drecv / dt * u.byte / u.s}



from typing import Optional, TypeVar, overload

from ..types import SupportsComparison

T = TypeVar("T", bound=SupportsComparison)


@overload
def clip(value: T, minimum: T, maximum: T, /) -> T: ...


@overload
def clip(value: T, absmax: T, /) -> T: ...


def clip(value: T, minimum: Optional[T] = None, maximum: Optional[T] = None, /) -> T:
    """Limit the ``value`` to the range [``minimum``, ``maximum``].

    Parameters
    ----------
    value
        Arbitrary parameter.
    minimum
        Lower limit of ``value``, if ``maximum`` is given. Otherwise, absolute value of
        this is used as the upper limit, and negative of that is the lower limit.
    maximum
        Upper limit of ``value``.

    Examples
    --------
    >>> neclib.utils.clip(1.2, 0, 1)
    1
    >>> neclib.utils.clip(41, 0, 100)
    41
    >>> neclib.utils.clip(-4, 3)
    -3

    """
    if minimum is None:
        raise ValueError("Bounding value required.")
    _max = maximum if maximum is not None else abs(minimum)  # type: ignore
    _min = minimum if maximum is not None else -1 * _max  # type: ignore
    if _min > _max:
        raise ValueError("Minimum should be less than maximum.")
    return min(max(_min, value), _max)  # type: ignore



from abc import abstractmethod

import astropy.units as u

from ..device_base import DeviceBase


class PowerMeter(DeviceBase):
    @abstractmethod
    def get_power(self) -> u.Quantity: ...

    def close(self) -> None: ...



import time
from typing import Optional, Union

import astropy.units as u
import ogameasure

from ... import get_logger
from ...core.security import busy
from ...core.units import dBm
from ...utils import skip_on_simulator
from .signal_generator_base import SignalGenerator


class E8257D(SignalGenerator):
    """Signal Generator, which can supply Local Signal.

    Notes
    -----

    Configuration items for this device:

    communicator : str
        Communicator of thermometer. GPIB or LAN can be chosen.

    host : str
        IP address for GPIB and ethernet communicator.

    gpib_port : int
        GPIB port of using devices. Please check device setting.
        If you use GPIB communicator, you must set this parameter.

    lan_port : int
        LAN port of using devices. This parameter is setted to 5025 by manufacturer.
        If you use LAN communicator, you must set this parameter.

    """

    Manufacturer: str = "Agilent"
    Model = "E8257D"

    Identifier = "host"

    @skip_on_simulator
    def __init__(self) -> None:
        self.logger = get_logger(self.__class__.__name__)

        if self.Config.communicator == "GPIB":
            com = ogameasure.gpib_prologix(self.Config.host, self.Config.gpib_port)
        elif self.Config.communicator == "LAN":
            com = ogameasure.ethernet(self.Config.host, self.Config.lan_port)
        else:
            self.logger.warning(
                f"There is not exsited communicator: {self.Config.communicator}."
                "Please choose USB or GPIB."
            )
        self.sg = ogameasure.Agilent.E8257D(com)

    def set_freq(self, GHz: Union[int, float]) -> None:
        with busy(self, "busy"):
            self.sg.freq_set(GHz)
            time.sleep(0.01)
            return

    def set_power(self, dBm: Union[int, float]) -> None:
        with busy(self, "busy"):
            self.sg.power_set(dBm)
            time.sleep(0.01)
            return

    def get_freq(self) -> u.Quantity:
        with busy(self, "busy"):
            f = self.sg.freq_query()
            time.sleep(0.01)
            return f * u.Hz

    def get_power(self) -> u.Quantity:
        with busy(self, "busy"):
            f = self.sg.power_query()
            time.sleep(0.01)
            return f * dBm

    def start_output(self) -> None:
        with busy(self, "busy"):
            self.sg.output_on()
            time.sleep(0.01)
            return

    def stop_output(self) -> None:
        with busy(self, "busy"):
            self.sg.output_off()
            time.sleep(0.01)
            return

    def get_output_status(self) -> Optional[bool]:
        with busy(self, "busy"):
            f = self.sg.output_query()
            time.sleep(0.01)
            if f == 1:
                return True
            elif f == 0:
                return False
            else:
                return None

    def finalize(self) -> None:
        self.stop_output()
        self.sg.com.close()

    def close(self) -> None:
        self.sg.com.close()



from typing import Tuple

import astropy.units as u
import numpy as np
import scipy

from .pointing_error import PointingError


class OMU1P85M(PointingError):
    def apply_offset(
        self, az: u.Quantity, el: u.Quantity
    ) -> Tuple[u.Quantity, u.Quantity]:
        dx = (
            self.a1 * np.sin(el)
            + self.a2
            + self.a3 * np.cos(el)
            + self.b1 * np.sin(az) * np.sin(el)
            - self.b2 * np.cos(az) * np.sin(el)
            + self.c1 * np.sin(az - el)
            + self.c2 * np.cos(az - el)
            + self.d1
            + self.e1 * np.cos(el)
            - self.e2 * np.sin(el)
        )
        dAz = dx / np.cos(el)
        dy = (
            self.b1 * np.cos(az)
            + self.b2 * np.sin(az)
            + self.b3
            + self.g1 * el
            + self.c1 * np.cos(az - el)
            - self.c2 * np.sin(az - el)
            + self.d2
            + self.e1 * np.sin(el)
            + self.e2 * np.cos(el)
        )
        dEl = dy

        # The above is defined as (refracted + offset = apparent), so reverse the sign
        return az + dAz, el + dEl

    def apply_inverse_offset(
        self, az: u.Quantity, el: u.Quantity
    ) -> Tuple[u.Quantity, u.Quantity]:
        def res(x):
            Az0 = az.value
            El0 = el.value
            a1 = self.a1.deg
            a2 = self.a2.deg
            a3 = self.a3.deg
            b1 = self.b1.deg
            b2 = self.b2.deg
            b3 = self.b3.deg
            c1 = self.c1.deg
            c2 = self.c2.deg
            d1 = self.d1.deg
            d2 = self.d2.deg
            g1 = self.g1
            e1 = self.e1.deg
            e2 = self.e2.deg
            dx = (
                (a3 + e1) * np.cos(x[1] * np.pi / 180)
                + (a1 - e2) * np.sin(x[1] * np.pi / 180)
                + (b1 + c2) * np.sin(x[0] * np.pi / 180) * np.sin(x[1] * np.pi / 180)
                - (b2 + c1) * np.cos(x[0] * np.pi / 180) * np.sin(x[1] * np.pi / 180)
                + c1 * np.sin(x[0] * np.pi / 180) * np.cos(x[0] * np.pi / 180)
                + c2 * np.cos(x[0] * np.pi / 180) * np.cos(x[1] * np.pi / 180)
                + a2
                + d1
            )

            dy = (
                e2 * np.cos(x[1] * np.pi / 180)
                + e1 * np.sin(x[1] * np.pi / 180)
                + b1 * np.cos(x[0] * np.pi / 180)
                + b2 * np.sin(x[0] * np.pi / 180)
                + c1 * np.cos(x[0] * np.pi / 180) * np.cos(x[1] * np.pi / 180)
                + c1 * np.sin(x[0] * np.pi / 180) * np.sin(x[1] * np.pi / 180)
                - c2 * np.sin(x[0] * np.pi / 180) * np.cos(x[1] * np.pi / 180)
                + c2 * np.cos(x[0] * np.pi / 180) * np.sin(x[1] * np.pi / 180)
                + g1 * x[1]
                + b3
                + d2
            )

            f = np.zeros(2, dtype=float)

            f[0] = (x[0] + (dx / np.cos(x[1] * np.pi / 180)) - Az0) * (
                1
                + (1 / np.cos(x[1] * np.pi / 180))
                * (
                    (b1 + c2) * np.cos(x[0] * np.pi / 180) * np.sin(x[1] * np.pi / 180)
                    + (b2 + c1)
                    * np.sin(x[0] * np.pi / 180)
                    * np.sin(x[1] * np.pi / 180)
                    + c1 * np.cos(x[0] * np.pi / 180) * np.cos(x[1] * np.pi / 180)
                    - c2 * np.sin(x[0] * np.pi / 180) * np.cos(x[1] * np.pi / 180)
                )
            ) + (x[1] + dy - El0) * (
                -b1 * np.sin(x[0] * np.pi / 180)
                + b2 * np.cos(x[0] * np.pi / 180)
                - c1 * np.sin(x[0] * np.pi / 180) * np.cos(x[1] * np.pi / 180)
                + c1 * np.cos(x[0] * np.pi / 180) * np.sin(x[1] * np.pi / 180)
                - c2 * np.cos(x[0] * np.pi / 180) * np.cos(x[1] * np.pi / 180)
                - c2 * np.sin(x[0] * np.pi / 180) * np.sin(x[1] * np.pi / 180)
            )
            f[1] = (x[0] + (dx / np.cos(x[1] * np.pi / 180)) - Az0) * (
                np.tan(x[1] * np.pi / 180) * dx
                + (1 / np.cos(x[1] * np.pi / 180))
                * (
                    -(a3 + e1) * np.sin(x[1] * np.pi / 180)
                    + (a1 - e2) * np.cos(x[1] * np.pi / 180)
                    + (b1 + c2)
                    * np.sin(x[0] * np.pi / 180)
                    * np.cos(x[1] * np.pi / 180)
                    - (b2 + c1)
                    * np.cos(x[0] * np.pi / 180)
                    * np.cos(x[1] * np.pi / 180)
                    - c1 * np.sin(x[0] * np.pi / 180) * np.sin(x[1] * np.pi / 180)
                    - c2 * np.cos(x[0] * np.pi / 180) * np.sin(x[1] * np.pi / 180)
                )
            ) + (x[1] + dy - El0) * (
                -e2 * np.sin(x[1] * np.pi / 180)
                + e1 * np.cos(x[1] * np.pi / 180)
                - c1 * np.sin(x[0] * np.pi / 180) * np.sin(x[1] * np.pi / 180)
                + c1 * np.sin(x[0] * np.pi / 180) * np.cos(x[1] * np.pi / 180)
                + c2 * np.sin(x[0] * np.pi / 180) * np.sin(x[1] * np.pi / 180)
                + c2 * np.cos(x[0] * np.pi / 180) * np.cos(x[1] * np.pi / 180)
                + g1
                + 1
            )
            return f

        az0, el0 = self.apply_offset(az, el)
        x0 = np.array([az0.deg, el0.deg])

        ans = scipy.optimize.root(res, x0, method="hybr", tol=1e-13)
        az, el = ans.x

        return az * u.deg, el * u.deg

    def fit(self, *args, **kwargs):
        raise NotImplementedError("Fitting is not implemented for this model.")



"""Device controllers.

NECLIB highly abstracts the devices to their *kind*, say "motor", "spectrometer", and so
on. This will allow users to write a code without model-specific configuration process
and even complete reuse of the code on device replacement. The model-specific
configurations should be stored in ``neclib.config`` object.

"""

import importlib
import sys
from pathlib import Path
from types import ModuleType
from typing import Dict, Type

from ..core import get_logger
from . import selector
from .device_base import DeviceBase, Devices

# Warn system-dependent implementations
if not sys.platform.startswith("linux"):
    logger = get_logger(__file__)
    logger.warning(
        "Device drivers for Interface PCI boards are only supported on Linux."
    )

# Find and import all device controller implementations, for subclass search in
# `device_base.get_device_list()`
paths = Path(__file__).parent.iterdir()
module_paths = filter(lambda p: p.is_dir() and p.name[0] not in "._", paths)
impl_modules = [
    importlib.import_module(f".{m.name}", __package__) for m in module_paths
]

implementations: Dict[str, Type[DeviceBase]]
"""List of all available implementations."""

parsed: Dict[str, Devices] = {}
"""List of parsed device implementations."""

here = sys.modules[__name__]


def reload():
    global implementations, parsed
    [delattr(here, k) for k in parsed.keys()]
    implementations = selector.list_implementations()
    parsed = selector.parse_device_configuration()
    for k, v in parsed.items():
        if (not hasattr(here, k)) or isinstance(getattr(here, k), ModuleType):
            setattr(here, k, v)
        else:
            logger = get_logger(__file__)
            logger.warning(f"Cannot set name {k!r}; already defined in this module.")


reload()



import queue
import socket
import struct
import time
import traceback
from threading import Event, Thread
from typing import Dict, List, Tuple

from ... import get_logger
from .spectrometer_base import Spectrometer


class AC240(Spectrometer):
    Manufacturer = "Agilent"
    Model: str = "AC240"
    Identifier = "host"

    def __init__(self) -> None:
        self.logger = get_logger(self.__class__.__name__)

        self.s = socket.socket()
        self.s.connect((self.Config.host, self.Config.port))
        self.msg_fmt = self.Config.msg_fmt
        self.msg_size = struct.calcsize(self.msg_fmt)
        self.board_id = self.Config.board_id
        self.data_queue = queue.Queue(maxsize=self.Config.record_quesize)
        self.thread = None
        self.event = None
        self.start()

    def start(self) -> None:
        if (self.thread is not None) or (self.event is not None):
            self.stop()
        self.thread = Thread(target=self._read_data, daemon=True)
        self.event = Event()
        self.thread.start()

    def receive(self):
        received = 0
        d = b""

        while received != self.msg_size:
            d += self.s.recv(self.msg_size - received)
            received = len(d)
            continue

        d = self.msg_unpack(d)

        return d

    def msg_unpack(self, buff):
        ud = struct.unpack(self.msg_fmt, buff)
        dd = {
            "timestamp": ud[0],
            "spectrum": ud[1:16385],
            "total_power": ud[16385 + 0],
            "integ_time": ud[16385 + 1],
            "temp_board": int(ud[16385 + 2]),
            "temp_fpga": ud[16385 + 3],
            "overflow_fpga": int(ud[16385 + 4]),
            "overflow_ad": ud[16385 + 5],
        }
        return dd

    def _read_data(self) -> None:
        while (self.event is not None) and (not self.event.is_set()):
            if self.data_queue.full():
                if self.warn:
                    self.logger.warning(
                        "Dropping the data due to low readout frequency."
                    )
                    self.warn = False
                self.data_queue.get()

            try:
                data = self.receive()
                self.data_queue.put((time.time(), {self.board_id: data["spectrum"]}))

            except struct.error:
                exc = traceback.format_exc()
                self.logger.warning(exc[slice(0, min(len(exc), 100))])

    def get_spectra(self) -> Tuple[float, Dict[int, List[float]]]:
        self.warn = True
        return self.data_queue.get()

    def stop(self) -> None:
        if self.event is not None:
            self.event.set()
        if self.thread is not None:
            self.thread.join()
        self.event = self.thread = None

    def finalize(self):
        self.stop()
        self.s.close()

    def change_spec_ch(self, chan):
        self.logger.warning("AC240 have not implemented binning mode yet.")
        pass



__all__ = ["CPZ6204"]

from astropy import units as u

from ... import get_logger, utils
from .encoder_base import Encoder


class CPZ6204(Encoder):
    """Encoder readout.

    Notes
    -----
    Configuration items for this device:

    rsw_id : {0, 1, ..., 16} or {"0", "1", ..., "9", "A", ..., "F"}
        Board identifier. This should be set to the same value as the rotary switch
        "RSW1" mounted on the side of the board. The board is shipped with default
        RSW1 setting of 0. This ID would be non-zero, when multiple PCI board of same
        model are mounted on a single FA (Factory Automation) controller.

    See defaults setting file in neclib/defaults/config.toml.
    """

    Manufacturer = "Interface"
    Model = "CPZ6204"

    Identifier = "rsw_id"

    def __init__(self) -> None:
        self.logger = get_logger(self.__class__.__name__)
        self.rsw_id = self.Config.rsw_id
        self.separation = self.Config.separation

        self.resolution = 360 * 3600 / (23600 * 400)

        if self.separation == "antenna":
            self.io = self._antenna_initialize()
        elif self.separation == "dome":
            self.io = self._dome_initialize()
        else:
            raise ValueError("Can't initialize. Please check separation in config")

    @utils.skip_on_simulator
    def _antenna_initialize(self):
        import pyinterface

        self.az_adjust = self.Config.az_adjust
        self.el_adjust = self.Config.el_adjust

        io = pyinterface.open(6204, self.rsw_id)
        if io is None:
            raise RuntimeError("Cannot communicate with the CPZ board.")
        mode = io.get_mode()
        if mode["mode"] == "":
            io.initialize()
            io.set_mode(mode="MD0 SEL1", direction=1, equal=0, latch=0, ch=1)
            io.set_mode(mode="MD0 SEL1", direction=1, equal=0, latch=0, ch=2)
            self.board_setting(io)
        else:
            pass
        return io

    @utils.skip_on_simulator
    def _dome_initialize(self):
        import pyinterface

        self.dome_adjust = self.Config.dome_adjust

        io = pyinterface.open(6204, self.rsw_id)
        if io is None:
            raise RuntimeError("Cannot communicate with the CPZ board.")
        # io.reset(ch=1)
        mode = io.get_mode()
        if mode["mode"] == "":
            io.set_mode("MD0", 0, 1, 0, ch=1)

        self.touchsensor_pos = [
            -391,
            -586,
            -780,
            -974,
            -1168,
            -1363,
            -1561,
            -1755,
            -1948,
            -2143,
            0,
            -197,
        ]
        self.dome_encoffset = self.Config.dome_encoffset
        self.dome_enc1loop = self.Config.dome_enc1loop
        self.dome_enc2arcsec = 3600.0 * 360 / self.dome_enc1loop
        self.dome_enc_tel_offset = self.Config.dome_enc_tel_offset * 360
        return io

    def get_dome_reading(self):
        # counter = self.dio.get_position()
        counter = self.io.get_counter(ch=1)
        # print('self,dio.get_counter : ', counter.to_int())
        counter = int(counter)  # .to_int()
        dome_enc_arcsec = -int(
            ((counter - self.dome_encoffset) * self.dome_enc2arcsec)
            - self.dome_enc_tel_offset
        )
        while dome_enc_arcsec > 1800.0 * 360:
            dome_enc_arcsec -= 3600.0 * 360
        while dome_enc_arcsec <= -1800.0 * 360:
            dome_enc_arcsec += 3600.0 * 360
        dome_position = dome_enc_arcsec / 3600
        self.dome_position = dome_position * u.deg
        return self.dome_position

    def get_reading(self):
        cntAz = int(self.io.get_counter(unsigned=False, ch=1))
        cntEl = int(self.io.get_counter(unsigned=False, ch=2))
        """unsigned
        if cntAz < 360*3600./self.resolution:
            #encAz = (324*cntAz+295)/590
            encAz = cntAz*self.resolution
        else:
            encAz = -(2**32-cntAz)*self.resolution
            pass
        """
        encAz = cntAz * self.resolution
        Az = encAz / 3600
        Az = Az * u.deg

        """ unsigned
        if cntEl < 360*3600./self.resolution:
            #encEl = (324*cntEl+295)/590
            encEl = cntEl*self.resolution
        else:
            encEl = -(2**32-cntEl)*self.resolution
            pass
        """
        encEl = cntEl * self.resolution
        _El = (encEl / 3600) + self.el_adjust
        El = _El * u.deg
        AzEl = {"az": Az, "el": El}

        return AzEl  # , _utc]

    def dome_set_counter(self, limit):
        # self.dio.ctrl.set_counter(counter)
        counter = self.touchsensor_pos[limit - 1] + self.dome_encoffset
        self.io.set_counter(counter, ch=1)
        return

    def finalize(self) -> None:
        pass

    def board_setting(self, io, z_mode=""):
        self.logger.info("Initialize start")
        io.set_z_mode(clear_condition=z_mode, latch_condition="", z_polarity=0, ch=1)
        io.set_z_mode(clear_condition=z_mode, latch_condition="", z_polarity=0, ch=2)
        print("origin setting mode : ", z_mode)
        self.logger.info("initialize end")
        return



__all__ = ["AZD_AD"]
import time
from typing import Union

import ogameasure

from ... import get_logger
from ...core.security import busy
from .motor_base import Motor


class AZD_AD(Motor):
    """Motor Controller for device of Oriental Motor

    Notes
    -----

    This code is used for direct operating of Oriental Motor Driver
    by using network converter (RS485 <-> Ethernet).
    For wiring and setup, please check the manual of this device.

    Configuration items for this device:

    host : str
        IP address for ethernet communicator.
        If you use LAN communicator, you must set this parameter.

    port : int
        Ethernet port of using devices. Please check device setting.
        If you use LAN communicator, you must set this parameter.

    position :Dict[str, int]
        Human-readable Position and step dictionary. The value should
        be mapping from human readable version (str) to device level
        identifier (int). You can assign any name to the step.
        For example: {insert = 8000, remove = 20000}

    velocity : int
        Motor velocity.
    low_limit : int
        Lower limit of motor motion.
    high_limit : int
        Higher limit of motor motion.
    """

    Model = "AZD_AD"
    Manufacturer = "OrientalMotor"

    Identifier = "host"

    def __init__(self):
        self.logger = get_logger(self.__class__.__name__)

        host = self.Config.host
        port = self.Config.port
        com = ogameasure.ethernet(host, port)
        self.motor = ogameasure.OrientalMotor.azd_ad(com)
        self.motor.initialize()
        self.velocity = self.Config.velocity

    def set_step(self, step: Union[int, str], axis=None) -> None:
        if isinstance(step, str):
            step = self.Config.position[step.lower()]
        if (step >= self.Config.low_limit) & (step <= self.Config.high_limit):
            self.motor.direct_operation(location=step, speed=self.velocity)
        else:
            raise ValueError(f"Over limit range: {step}")
        return

    def set_speed(self, speed=None, axis=None) -> None:
        raise RuntimeError(
            "This device don't have linear-uniform-motion mode."
            "Please use `set_step` function."
        )

    def get_step(self, axis=None) -> int:
        with busy(self, "busy"):
            step = self.motor.get_current_position()
            return step

    def get_speed(self, axis=None) -> None:
        raise NotImplementedError(
            "This device don't have linear-uniform-motion mode."
            "Please use `get_step` function."
        )

    def move_home(self) -> None:
        self.motor.zero_return()
        return

    def get_alarm(self) -> str:
        alarm = self.motor.alarm_query()
        return alarm

    def reset_alarm(self) -> None:
        self.motor.alarm_reset()
        return

    def finalize(self) -> None:
        self.motor.zero_return()
        time.sleep(5)
        self.motor.com.close()
        return

    def close(self) -> None:
        self.motor.com.close()



__all__ = ["CPZ7204"]

import time

import astropy.units as u

from ... import get_logger, utils
from .motor_base import Motor


class CPZ7204(Motor):
    """Pulse motion controller for  NANTEN2 M4.

    Notes
    -----
    Configuration items for this device:

    rsw_id : {0, 1, ..., 16} or {"0", "1", ..., "9", "A", ..., "F"}
        Board identifier. This should be set to the same value as the rotary switch
        "RSW1" mounted on the side of the board. The board is shipped with default
        RSW1 setting of 0. This ID would be non-zero, when multiple PCI board of same
        model are mounted on a single FA (Factory Automation) controller.

    See defaults setting file in neclib/defaults/config.toml.
    """

    Manufacturer = "Interface"
    Model = "CPZ7204"

    Identifier = "rsw_id"

    def __init__(self) -> None:
        self.logger = get_logger(self.__class__.__name__)
        self.rsw_id = self.Config.rsw_id
        self.io = self._initialize_io()

    @utils.skip_on_simulator
    def _initialize_io(self):
        import pyinterface

        io = pyinterface.open(7204, self.rsw_id)
        if io is None:
            raise RuntimeError("Cannot communicate with the PCI board.")
        io.initialize()

        return io

    def set_step(self, position: str) -> None:
        self.move(position)
        return

    def get_step(self) -> str:
        status = self.io.get_status()
        if status["busy"]:
            position = "MOVE"
        elif status["limit"]["+EL"] == 0:
            position = "OUT"
        elif status["limit"]["-EL"] == 0:
            position = "IN"
        else:
            self.logger.warning("limit error")
            position = "ERROR"
        return position

    def move(self, position: str) -> None:
        if position == "OUT":
            step = 1
            self.logger.info("m4 move out")
        elif position == "IN":
            step = -1
            self.logger.info("m4 move in")
        else:
            self.logger.warning("parameter error")
            return
        self.io.set_motion(
            mode="JOG",
            acc_mode="SIN",
            low_speed=100,
            speed=1000,
            acc=500,
            step=step,
            axis=1,
        )
        self.io.start_motion(mode="JOG")
        self.logger.info("start_motion")
        time.sleep(0.5)
        return

    def stop(self, mode: str) -> None:
        # mode: "DEC", "IMMEDIATE"
        self.io.stop_motion(mode, 1)
        return

    def finalize(self) -> None:
        self.io.output_do([0, 0, 0, 0], 1)

    def set_speed(self, speed: float, axis: str) -> None:
        pass

    def get_speed(self, axis: str) -> u.Quantity:
        pass



__all__ = ["ND287"]

import astropy.units as u
import ogameasure

from ... import utils
from ...core.security import busy
from .encoder_base import Encoder


class ND287(Encoder):
    """Encoder readout.

    Notes
    -----

    Configuration items for this device:

    port : int
        USB port of using devices.

    """

    Manufacturer = "HEIDENHAIN"
    Model = "ND287"

    Identifier = "port"

    @utils.skip_on_simulator
    def __init__(self) -> None:
        self.io = ogameasure.HEIDENHAIN.ND287(self.Config.port)

    def get_reading(self) -> u.Quantity:
        with busy(self, "busy"):
            raw = self.io.output_position_display_value()
        return float(raw.strip(b"\x02\x00\r\n").decode()) << u.deg  # type: ignore

    def finalize(self) -> None:
        try:
            self.io._enc.close()
        except AttributeError:
            pass



import os
import re
import warnings
from dataclasses import dataclass
from itertools import chain
from typing import (
    IO,
    Any,
    Callable,
    Dict,
    Generic,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from tomlkit.container import Container
from tomlkit.items import Item

from ..exceptions import NECSTAccessibilityWarning, NECSTParameterNameError
from ..files import toml
from ..formatting.dict_to_html import html_repr_of_dict
from .parameters import Parameters

T_value = TypeVar("T_value")


@dataclass
class _RichParameter(Generic[T_value]):
    key: str
    value: T_value

    @property
    def parsed(self) -> Any:
        """Value which may be converted by attached parser."""
        if not hasattr(self, "_parsed_value"):
            _value = self.value
            if isinstance(_value, (Container, Item)):
                # TOML items will pass `isinstance()` checks but will fail `type()`
                # checks. Some scripts checks the type by `type()`, so we need to
                # convert them to corresponding Python built-in types.
                _value = _value.unwrap()
            self._parsed_value = self.parser(_value)
        return self._parsed_value

    @property
    def parser(self) -> Callable[[T_value], Any]:
        """Parser to convert value to another type or modify the value."""
        return self._parser if hasattr(self, "_parser") else lambda x: x

    @parser.setter
    def parser(self, _parser: Callable[[T_value], Any]) -> None:
        if hasattr(self, "_parsed_value"):
            # ``parsed`` property reuses the parsed value, so when the parser possibly
            # changed, the parsed value should be removed.
            del self._parsed_value
        self._parser = _parser

    def __str__(self) -> str:
        return str(self.parsed)


class RichParameters(Parameters):
    """Parameters with flexible look-up and high-level representation support.

    Parameters are not always just a key-value pair. Sometimes, we need higher level
    representation; all-in-one object that has rich information and methods about the
    parameter. This class provides a way to store such parameters.

    Parameters
    ----------
    _prefix
        Internally used prefix, which enables name-based filtering of parameters.
    **kwargs
        Parameters to be stored.

    Examples
    --------
    >>> params = RichParameters(a=1, b=2, **{"c[deg]": 3})

    You can attach a parser to a parameter, which will be used to convert the value to
    another high-level representation or modify the value.

    >>> params.attach_parser(b=lambda x: x * 2 * u.m)
    >>> params.b
    <Quantity 4. m>

    You can filter parameters by name, using the parameter prefix to accessing the
    attribute.

    >>> params = RichParameters(
    ...     general_param1=1,
    ...     specific_param1=2,
    ...     **{"general_param2[deg]": 3}
    ... )
    >>> params.general
    RichParameters
    general_param1 = 1 (_RichParameter)
    general_param2 = 3d00m00s (_RichParameter)

    """

    __slots__ = ("_prefix",)

    # Dot is used as namespace separator, so it's valid letter for parameter with
    # flexible look-up support
    _unit_matcher = re.compile(r"([\w\.]*)\[([\w/\s\*\^-]*)\]")

    _view_class: Optional[Type["RichParameters"]] = None

    def __init__(self, _prefix: str = "", /, **kwargs: Any) -> None:
        self._prefix = _prefix
        params = {
            k: v if isinstance(v, _RichParameter) else _RichParameter(k, v)
            for k, v in kwargs.items()
        }
        with warnings.catch_warnings():
            # The use of dots are valid in this class
            warnings.simplefilter("ignore", category=NECSTAccessibilityWarning)
            super().__init__(**params)
        self._parameters: Dict[str, _RichParameter[Any]]

        # Default aliases, which enable attribute access
        aliases = {k.replace(".", "_"): k for k in self._parameters.keys() if "." in k}
        self.attach_aliases(**aliases)

    @classmethod
    def from_file(cls, file: Union[os.PathLike, str, IO], /):
        """Read parameters from a TOML file.

        Parameters
        ----------
        file
            The file path or file object.

        Examples
        --------
        >>> params = RichParameters.from_file("path/to/file.toml")

        You can also read parameters from a file-like object.

        >>> with open("path/to/file.toml") as f:
        ...     params = RichParameters.from_file(f)

        And if you wish, remote files can be read:

        >>> params = RichParameters.from_file("https://example.com/file.toml")

        """
        file_path = isinstance(file, (os.PathLike, str))
        _params = toml.read(file)
        params = toml.flatten(_params)

        inst = cls(**params)
        if file_path:
            inst._metadata["path"] = file
        return inst

    def _parse(self, k: str, v: _RichParameter) -> Tuple[str, _RichParameter]:
        parsed_k, parsed_v = super()._parse(k, v.value)
        v.key = parsed_k
        v.value = parsed_v
        return parsed_k, v

    def attach_parsers(self, **kwargs: Callable[[Any], Any]) -> None:
        """Attach parsers to parameters.

        Parameters
        ----------
        **kwargs
            Parsers to be attached to parameters.

        Raises
        ------
        NECSTParameterNameError
            No parameter with the given name (key of the ``kwargs``) exists.

        Examples
        --------
        >>> params = RichParameters(a=1, b=2, **{"c[deg]": 3})
        >>> params.attach_parsers(b=lambda x: x * 2 * u.m)
        >>> params.b
        <Quantity 4. m>

        """
        for k, v in kwargs.items():
            if k in self._parameters:
                self._parameters[k].parser = v
            elif k in self._aliases:
                self._parameters[self._aliases[k]].parser = v

            try:
                self._pick(k).parser = v
            except KeyError:
                raise NECSTParameterNameError(f"Unknown parameter: {k!r}")

    def get(self, key: str, /) -> Any:
        """Flexibly look-up the parameters.

        Implementation of the ``__getitem__`` and ``__getattr__`` special methods.

        Parameters
        ----------
        key
            The name of the parameter.

        """
        slots = chain.from_iterable(
            getattr(cls, "__slots__", []) for cls in self.__class__.__mro__
        )
        if key in slots:
            raise KeyError(key)
        if key == self._prefix:
            return self

        try:
            return self._pick(key).parsed
        except KeyError:
            filtered = self._filter(key)

        if filtered:
            prefix = self._prefix + "_" + key if self._prefix else key
            inst = (self._view_class or self.__class__)(prefix, **filtered)
            inst._metadata = self._metadata
            inst._metadata.update(parent=self)
            for k, v in self._aliases.items():
                if v in filtered:
                    inst._aliases[k] = v
            return inst
        raise KeyError(key)

    def __getitem__(self, key: str, /) -> Any:
        return self.get(key)

    def __getattr__(self, key: str, /) -> Any:
        try:
            return self.get(key)
        except KeyError as e:
            # Should raise AttributeError to deny permissiveness to nonexistent
            # attributes. If an object performs as if it has any attributes by not
            # raising AttributeError, Jupyter Notebook won't trust the result of
            # `hasattr(obj, "_repr_html_")` hence don't display rich representations.
            # See also: https://github.com/jupyter/notebook/issues/2014
            raise AttributeError(f"No attribute {key!r}") from e

    def _filter(self, key: str, /) -> Dict[str, _RichParameter]:
        """Extract parameters which key starts with given key."""
        extracted = {}
        qualkey = self._prefix + "_" + key if self._prefix else key
        for k, v in self._parameters.items():
            if self._match(v.key, qualkey):
                extracted[k] = v
        for k, v in self._aliases.items():
            if self._match(k, qualkey):
                extracted[v] = self._parameters[v]
        return extracted

    def _pick(self, key: str, /) -> _RichParameter:
        """Pick-up at most one parameter, which key exactly matches to given one."""
        qualkey = self._prefix + "_" + key if self._prefix else key
        for v in self._parameters.values():
            if self._match(qualkey, v.key, strict=True):
                return v
        for k, v in self._aliases.items():
            if self._match(qualkey, k, strict=True):
                return self._parameters[v]
        raise KeyError(key)

    def _normalize(self, qualkey: str, /) -> str:
        """Normalize the key to enable structure insensitive access.

        Here normalization means replacing ``.`` with ``_``. This will convert keys to
        names accessible as attributes.

        Parameters
        ----------
        qualkey
            The key to be normalized. It should be a qualified key, i.e. the key with
            the prefix.

        """
        return qualkey.replace(".", "_")

    def _match(self, k1: str, k2: str, /, *, strict: bool = False) -> bool:
        """Judge whether namespace ``k1`` is equal to or contains key ``k2``.

        Parameters
        ----------
        k1
            The namespace. Must be a qualified key.
        k2
            The key. Must be a qualified key.
        strict
            If ``True``, this method judges ``k1`` and ``k2`` are equal or not.
            Otherwise, ``k1`` contains ``k2`` or not.

        """
        if strict:
            return self._normalize(k1) == self._normalize(k2)
        return self._normalize(k1).startswith(self._normalize(k2))

    @property
    def parameters(self) -> Dict[str, Any]:
        """Return a copy of the raw parameters."""
        return {k: v.parsed for k, v in self._parameters.items()}

    def _repr_html_(self) -> str:
        """Rich representation for Jupyter Notebook."""
        return html_repr_of_dict(
            {k: v.parsed for k, v in self._parameters.items()},
            type(self),
            aliases=self._aliases,
            metadata=dict(**self._metadata, prefix=self._prefix),
        )



import astropy.units as u

from ...core.math import Random
from .vacuum_gauge_base import VacuumGauge


class VacuumGaugeSimulator(VacuumGauge):
    Manufacturer: str = ""
    Model: str = ""
    Identifier = ""
    is_simulator = True

    def __init__(self) -> None:
        self._rand = Random().walk(1e-8, 1e-9, -1)

    def get_pressure(self) -> u.Quantity:
        return 1e-8 * u.Torr  # type: ignore

    def finalize(self) -> None:
        pass

    def close(self) -> None:
        pass



"""Check user input, running environment and so on.

The implementations here will ensure the execution of NECST is safe thus reliable.
The `unsafe` may contain but not limited to the following situations:

* Malicious user input can be executed.
* Loss of realtime-ness in feedback control.

"""

from .busy_impl import busy  # noqa: F401
from .load_check import LoadChecker  # noqa: F401
from .sanitize_impl import sanitize  # noqa: F401
